#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>

struct proc_dir_entry * netifrx;
struct net_device * ethx;

int vdeinject_write_func(struct file* file, const char* buffer, unsigned long count, void* ptr)
{
    struct sk_buff * skbi;
    unsigned char * data;
    int err;
    unsigned long len = count;

    /* allocate an skb with enough space */
    skbi = alloc_skb(len, GFP_ATOMIC);
    skb_reserve(skbi, NET_IP_ALIGN);

    /*  fill it up */
    data = skb_put(skbi, len+100); /* fixme: 100 bytes padding for head and tail is arbitirary */
    skbi->csum = csum_and_copy_from_user(buffer, data, len, 0, &err);
    if (err)
    {
	kfree_skb(skbi);
	return -EFAULT;
    }

    /* tell it which device to send to */
    skbi->dev = ethx;
    skbi->input_dev = ethx;
    skbi->real_dev = ethx;

    /* queue it */
    netif_rx(skbi);

    return len;
}

int vdeinject_read_func(char* page, char** start, off_t off, int count, int* eof, void* ptr)
{
    return 0;
}

static int vdeinject_init(void)
{   
    printk(KERN_ALERT "Loading netifrx bridge driver\n");
    netifrx = create_proc_entry("vdeinject", 0200, NULL);
    if (netifrx == NULL)
    {
	 /* handle failure here */
	printk(KERN_ALERT "create_proc_entry failed\n");
	return 0;
    }
    netifrx->read_proc = vdeinject_read_func;
    netifrx->write_proc = vdeinject_write_func;
    netifrx->owner = THIS_MODULE;
    ethx = dev_get_by_name("eth0"); /* needs to be user configurable */
    if (ethx == NULL)
    {
	 /* handle failure here */
	printk(KERN_ALERT "get_dev_by_name on eth0 failed\n");
	remove_proc_entry("netifrx", NULL);
	return 0;
    }
    return 0;
}

static void vdeinject_exit(void)
{   
    printk(KERN_ALERT "Unloading netifrx bridge driver\n");
    remove_proc_entry("netifrx", NULL);
    __dev_put(ethx);
}

module_init(vdeinject_init);
module_exit(vdeinject_exit);

MODULE_LICENSE("GPL");

MODULE_AUTHOR("Jim Brown");           
MODULE_DESCRIPTION("netif_rx() bridge");

